#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import idaes.logger as idaeslog

from pyomo.environ import (
    ConcreteModel,
    units as pyunits,
    Expression,
    value,
    TransformationFactory,
)

from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    PhotothermalMembraneZO,
    CANDOPZO,
    PumpElectricityZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_results(m)

    add_costing(m)
    m.fs.costing.initialize()

    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)
    results = solve(m, checkpoint="solve flowsheet with costing")
    display_costing(m)

    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(
        solute_list=[
            "nitrogen",
            "phosphates",
            "bioconcentrated_phosphorous",
            "nitrous_oxide",
        ]
    )

    # unit models
    # feed
    m.fs.feed = FeedZO(property_package=m.fs.prop)

    # pump
    m.fs.pump = PumpElectricityZO(property_package=m.fs.prop, database=m.db)

    # photothermal membrane
    m.fs.photothermal = PhotothermalMembraneZO(
        property_package=m.fs.prop, database=m.db
    )

    # CANDO+P reactor
    # Note that CANDOPZO model electricity costing includes pumping costs for
    # the unit, so a pump has not been included in this flowsheet between the
    # photothermal membrane and CANDO+P units
    m.fs.candop = CANDOPZO(property_package=m.fs.prop, database=m.db)

    # product streams
    m.fs.photothermal_water = Product(property_package=m.fs.prop)
    m.fs.candop_treated = Product(property_package=m.fs.prop)
    m.fs.candop_byproduct = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.pump.inlet)
    m.fs.s02 = Arc(source=m.fs.pump.outlet, destination=m.fs.photothermal.inlet)
    m.fs.s03 = Arc(
        source=m.fs.photothermal.byproduct, destination=m.fs.photothermal_water.inlet
    )
    m.fs.s04 = Arc(source=m.fs.photothermal.treated, destination=m.fs.candop.inlet)
    m.fs.s05 = Arc(
        source=m.fs.candop.byproduct, destination=m.fs.candop_byproduct.inlet
    )
    m.fs.s06 = Arc(source=m.fs.candop.treated, destination=m.fs.candop_treated.inlet)

    # expand arcs
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_vol = 800 * pyunits.m**3 / pyunits.hr
    conc_mass_nitrogen = 35 * pyunits.mg / pyunits.liter
    conc_mass_phosphates = 6 * pyunits.mg / pyunits.liter
    conc_mass_bcp = 1e-6 * pyunits.mg / pyunits.liter
    conc_mass_no2 = 1e-6 * pyunits.mg / pyunits.liter

    m.fs.feed.flow_vol[0].fix(flow_vol)
    m.fs.feed.conc_mass_comp[0, "nitrogen"].fix(conc_mass_nitrogen)
    m.fs.feed.conc_mass_comp[0, "phosphates"].fix(conc_mass_phosphates)
    m.fs.feed.conc_mass_comp[0, "bioconcentrated_phosphorous"].fix(conc_mass_bcp)
    m.fs.feed.conc_mass_comp[0, "nitrous_oxide"].fix(conc_mass_no2)
    solve(m.fs.feed, checkpoint="solve feed block")

    # pump
    m.fs.pump.load_parameters_from_database(use_default_removal=True)
    m.fs.pump.lift_height.fix(4.2)  # head of 4.2 m

    # photothermal membrane
    m.fs.photothermal.load_parameters_from_database(use_default_removal=True)

    # CANDO+P reactor
    m.fs.candop.load_parameters_from_database(use_default_removal=True)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_results(m):
    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    print("++++++++++++++++++++ DISPLAY RESULTS ++++++++++++++++++++")
    print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")

    unit_list = ["pump", "photothermal", "candop"]
    for u in unit_list:
        m.fs.component(u).report()

    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")


def add_costing(m):
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "amo_1595_case_study.yaml",
    )

    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    costing_kwargs = {"flowsheet_costing_block": m.fs.costing}
    m.fs.pump.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.photothermal.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.candop.costing = UnitModelCostingBlock(**costing_kwargs)

    m.fs.costing.cost_process()

    m.fs.costing.add_electricity_intensity(m.fs.feed.properties[0].flow_vol)

    m.fs.costing.annual_water_inlet = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.bcp_recovery_mass = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.candop_treated.flow_mass_comp[0, "bioconcentrated_phosphorous"],
            to_units=pyunits.kg / m.fs.costing.base_period,
        ),
        doc="Mass of BCP generated per year",
    )

    m.fs.costing.value_bcp_recovery = Expression(
        expr=pyunits.convert(
            m.fs.costing.bcp_recovery_mass * m.fs.costing.bcp_cost,
            to_units=m.fs.costing.base_currency / m.fs.costing.base_period,
        ),
        doc="Value of BCP generated per year",
    )

    m.fs.costing.water_byproduct_volume = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.photothermal_water.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        ),
        doc="Volume of byproduct water generated per year",
    )

    m.fs.costing.value_water_byproduct = Expression(
        expr=pyunits.convert(
            m.fs.costing.water_byproduct_volume * m.fs.costing.water_cost,
            to_units=m.fs.costing.base_currency / m.fs.costing.base_period,
        ),
        doc="Value of byproduct water generated per year",
    )

    m.fs.costing.N2O_byproduct_mass = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.candop_byproduct.flow_mass_comp[0, "nitrous_oxide"],
            to_units=pyunits.kg / m.fs.costing.base_period,
        ),
        doc="Mass of byproduct nitrous oxide generated per year",
    )

    m.fs.costing.value_N2O_byproduct = Expression(
        expr=pyunits.convert(
            m.fs.costing.N2O_byproduct_mass * m.fs.costing.nitrous_oxide_cost,
            to_units=m.fs.costing.base_currency / m.fs.costing.base_period,
        ),
        doc="Value of byproduct nitrous oxide generated per year",
    )

    m.fs.costing.LCOW = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            pyunits.convert(
                m.fs.photothermal_water.properties[0].flow_vol,
                to_units=pyunits.m**3 / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of Water",
    )

    m.fs.costing.LCOW_with_revenue = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
            - m.fs.costing.value_bcp_recovery
            - m.fs.costing.value_N2O_byproduct
        )
        / (
            pyunits.convert(
                m.fs.photothermal_water.properties[0].flow_vol,
                to_units=pyunits.m**3 / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of Water including revenue from BCP and N20",
    )

    m.fs.costing.LCOT = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            pyunits.convert(
                m.fs.feed.properties[0].flow_vol,
                to_units=pyunits.m**3 / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of Treatment with respect to influent flowrate",
    )

    m.fs.costing.LCOT_with_revenue = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
            - m.fs.costing.value_bcp_recovery
            - m.fs.costing.value_N2O_byproduct
            - m.fs.costing.value_water_byproduct
        )
        / (
            pyunits.convert(
                m.fs.feed.properties[0].flow_vol,
                to_units=pyunits.m**3 / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of Treatment with respect to influent flowrate including revenue from BCP, N2O, and byproduct water",
    )

    m.fs.costing.LC_BCP = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            pyunits.convert(
                m.fs.candop_treated.properties[0].flow_mass_comp[
                    "bioconcentrated_phosphorous"
                ],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of BCP (not accounting for any revenue)",
    )

    m.fs.costing.LC_N2O = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            pyunits.convert(
                m.fs.candop_byproduct.properties[0].flow_mass_comp["nitrous_oxide"],
                to_units=pyunits.kg / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of N2O (not accounting for any revenue)",
    )


def display_costing(m):
    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    print("++++++++++++++++++++ DISPLAY COSTING ++++++++++++++++++++")
    print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")

    print("\n---------- Levelized costs ----------")
    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(f"Levelized Cost of Water: {LCOW:.3f} $/m^3 product water")

    LCOW_with_revenue = value(
        pyunits.convert(
            m.fs.costing.LCOW_with_revenue, to_units=pyunits.USD_2020 / pyunits.m**3
        )
    )
    print(
        f"Levelized Cost of Water including revenue: {LCOW_with_revenue:.3f} $/m^3 product water"
    )

    LCOT = value(
        pyunits.convert(m.fs.costing.LCOT, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(
        f"Levelized Cost of Treatment with respect to influent flowrate: {LCOT:.3f} $/m^3 feed water"
    )

    LCOT_with_revenue = value(
        pyunits.convert(
            m.fs.costing.LCOT_with_revenue, to_units=pyunits.USD_2020 / pyunits.m**3
        )
    )
    print(
        f"Levelized Cost of Treatment with respect to influent flowrate including revenue: {LCOT_with_revenue:.3f} $/m^3 feed water"
    )

    LC_BCP = value(
        pyunits.convert(m.fs.costing.LC_BCP, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of BCP (not including any revenue): {LC_BCP:.2f} $/kg")

    LC_N2O = value(
        pyunits.convert(m.fs.costing.LC_N2O, to_units=pyunits.USD_2020 / pyunits.kg)
    )
    print(f"Levelized Cost of N2O (not including any revenue): {LC_N2O:.2f} $/kg")

    print("\n------------- Capital costs -------------")
    DCC_normalized = value(
        pyunits.convert(
            (
                m.fs.pump.costing.direct_capital_cost
                + m.fs.photothermal.costing.direct_capital_cost
                + m.fs.candop.costing.direct_capital_cost
            )
            / m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.kUSD_2020 / (pyunits.m**3 / pyunits.hr),
        )
    )
    print(f"Normalized direct capital costs: {DCC_normalized:.2f} k$/(m^3/hr)")

    ICC_normalized = value(
        pyunits.convert(
            m.fs.costing.total_capital_cost / m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.kUSD_2020 / (pyunits.m**3 / pyunits.hr),
        )
    )
    print(f"Normalized total capital costs: {ICC_normalized:.2f} k$/(m^3/hr)")

    total_capital_cost = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.MUSD_2020)
    )
    print(f"Total Capital Costs: {total_capital_cost:.3f} M$")

    print(
        f"Pump capital cost: {value(pyunits.convert(m.fs.pump.costing.capital_cost, to_units=pyunits.kUSD_2020)):.2f} k$"
    )

    print(
        f"Photothermal membrane capital cost: {value(pyunits.convert(m.fs.photothermal.costing.capital_cost, to_units=pyunits.MUSD_2020)):.2f} M$"
    )

    print(
        f"CANDO+P capital cost: {value(pyunits.convert(m.fs.candop.costing.capital_cost, to_units=pyunits.MUSD_2020)):.2f} M$"
    )

    print("\n------------- Operating costs -------------")
    FMC_normalized = value(
        pyunits.convert(
            m.fs.costing.maintenance_cost / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.year,
        )
    )
    print(f"Normalized maintenance costs: {FMC_normalized:.2f} 1/year")

    OFOC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_fixed_operating_cost
            / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.year,
        )
    )
    print(f"Normalized other fixed operating cost: {OFOC_normalized:.2f} 1/year")

    EC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            * m.fs.costing.utilization_factor
            / m.fs.costing.annual_water_inlet,
            to_units=pyunits.USD_2020 / pyunits.m**3,
        )
    )
    print(f"Normalized electricity cost: {EC_normalized:.3f} $/m^3 of feed")

    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.electricity_intensity, to_units=pyunits.kWh / pyunits.m**3
        )
    )
    print(
        f"Electricity Intensity with respect to influent flowrate: {electricity_intensity:.3f} kWh/m^3"
    )

    total_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.kUSD_2020 / pyunits.year
        )
    )
    print(f"Total operating costs: {total_operating_costs:.2f} k$/year")

    fixed_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_fixed_operating_cost,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"Fixed operating costs: {fixed_operating_costs:.2f} k$/year")

    variable_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_variable_operating_cost,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"Variable operating costs: {variable_operating_costs:.2f} k$/year")

    electricity_operating_costs = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            * m.fs.costing.utilization_factor,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"Electricity operating costs: {electricity_operating_costs:.2f} k$/year")

    print("\n------------- Revenue -------------")

    BCP_revenue = value(
        pyunits.convert(
            m.fs.costing.value_bcp_recovery,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"BCP revenue: {BCP_revenue:.2f} k$/year")

    N2O_revenue = value(
        pyunits.convert(
            m.fs.costing.value_N2O_byproduct,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"N2O revenue: {N2O_revenue:.2f} k$/year")

    water_revenue = value(
        pyunits.convert(
            m.fs.costing.value_water_byproduct,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"Water revenue: {water_revenue:.2f} k$/year")

    total_revenue = value(
        pyunits.convert(
            m.fs.costing.value_bcp_recovery
            + m.fs.costing.value_N2O_byproduct
            + m.fs.costing.value_water_byproduct,
            to_units=pyunits.kUSD_2020 / pyunits.year,
        )
    )
    print(f"Total revenue: {total_revenue:.2f} k$/year")

    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import cost_by_flow_volume, register_costing_parameter_block


class EnergyRecoveryDeviceType(StrEnum):
    pressure_exchanger = "pressure_exchanger"


def cost_energy_recovery_device(
    blk,
    energy_recovery_device_type=EnergyRecoveryDeviceType.pressure_exchanger,
    cost_electricity_flow=True,
):
    """
    Energy recovery device costing method

    TODO: describe equations

    Args:
        energy_recovery_device_type: EnergyRecoveryDeviceType Enum indicating ERD type.
            Defaults to EnergyRecoveryDeviceType.pressure_exchanger.
        cost_electricity_flow: bool, if True, the ERD's work_mechanical will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    if energy_recovery_device_type == EnergyRecoveryDeviceType.pressure_exchanger:
        cost_pressure_exchanger_erd(blk)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for energy_recovery_device_type:"
            f" {energy_recovery_device_type}. Argument must be a member of the EnergyRecoveryDeviceType Enum."
        )


def build_energy_recovery_device_cost_param_block(blk):

    blk.pressure_exchanger_cost = pyo.Var(
        initialize=535,
        doc="Pressure exchanger cost",
        units=pyo.units.USD_2018 / (pyo.units.meter**3 / pyo.units.hours),
    )


@register_costing_parameter_block(
    build_rule=build_energy_recovery_device_cost_param_block,
    parameter_block_name="energy_recovery_device",
)
def cost_pressure_exchanger_erd(blk, cost_electricity_flow=True):
    """
    ERD pressure exchanger costing method

    TODO: describe equations

    Args:
        cost_electricity_flow (:obj:`bool`): if True, the ERD's work_mechanical will
            be converted to kW and costed as an electricity. Defaults to True.
    """
    t0 = blk.flowsheet().time.first()
    cost_by_flow_volume(
        blk,
        blk.costing_package.energy_recovery_device.pressure_exchanger_cost,
        pyo.units.convert(
            blk.unit_model.control_volume.properties_in[t0].flow_vol,
            (pyo.units.meter**3 / pyo.units.hours),
        ),
    )
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.work_mechanical[t0], to_units=pyo.units.kW
            ),
            "electricity",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)


def build_cstr_injection_cost_param_block(blk):
    # NOTE: costing data are from NREL Waste-to-Energy Model
    blk.capital_a_parameter = pyo.Var(
        initialize=1.1141e3,
        doc="A parameter for capital cost",
        units=pyo.units.USD_1997,
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=0.8324,
        doc="B parameter for capital cost",
        units=pyo.units.dimensionless,
    )


@register_costing_parameter_block(
    build_rule=build_cstr_injection_cost_param_block,
    parameter_block_name="cstr_injection",
)
def cost_cstr_injection(blk, cost_electricity_flow=True):
    """
    CSTR injection costing method
    """
    cost_cstr_injection_capital(
        blk,
        blk.costing_package.cstr_injection.capital_a_parameter,
        blk.costing_package.cstr_injection.capital_b_parameter,
    )

    t0 = blk.flowsheet().time.first()
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


def cost_cstr_injection_capital(blk, capital_a_parameter, capital_b_parameter):
    """
    Generic function for costing an CSTR injection system.
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    blk.capital_a_parameter = pyo.Expression(expr=capital_a_parameter)
    blk.capital_b_parameter = pyo.Expression(expr=capital_b_parameter)

    print(f"base_currency: {blk.costing_package.base_currency}")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.capital_a_parameter
            * (blk.unit_model.control_volume.volume[0] / pyo.units.m**3)
            ** blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a clarifier unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("ClarifierZO")
class ClarifierZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Clarifier unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "clarifier"

        build_sido(self)
        constant_intensity(self)

        if self.config.process_subtype == "HRCS_clarifier":

            self.ferric_chloride_dose = Var(
                self.flowsheet().time,
                units=pyunits.mg / pyunits.L,
                bounds=(0, None),
                doc="Dosing rate of ferric chloride",
            )
            self._fixed_perf_vars.append(self.ferric_chloride_dose)

            self.ferric_chloride_demand = Var(
                self.flowsheet().time,
                units=pyunits.kg / pyunits.hr,
                bounds=(0, None),
                doc="Consumption rate of ferric chloride",
            )
            self._perf_var_dict["Ferric Chloride Demand"] = self.ferric_chloride_demand

            @self.Constraint(
                self.flowsheet().time, doc="ferric chloride demand constraint"
            )
            def ferric_chloride_demand_equation(b, t):
                return b.ferric_chloride_demand[t] == pyunits.convert(
                    b.ferric_chloride_dose[t] * b.properties_in[t].flow_vol,
                    to_units=pyunits.kg / pyunits.hr,
                )

    @property
    def default_costing_method(self):
        return self.cost_clarifier

    @staticmethod
    def cost_clarifier(blk, number_of_parallel_units=1):
        """
        General method for costing clarifiers. Costing is carried out
        using either the general_power_law form or the standard form which
        computes HRT, sizing costs, and chemical input costs.
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units (default: 1)
        """
        # Get cost method for this technology
        cost_method = blk.unit_model._get_unit_cost_method(blk)
        valid_methods = ["cost_power_law_flow", "cost_HRCS_clarifier"]
        if cost_method == "cost_power_law_flow":
            blk.unit_model.cost_power_law_flow(blk, number_of_parallel_units)
        elif cost_method == "cost_HRCS_clarifier":
            # NOTE: number of units does not matter for cost_HRCS_clarifier
            #       as its a linear function of membrane area
            blk.unit_model.cost_HRCS_clarifier(blk)
        else:
            raise KeyError(
                f"{cost_method} is not a relevant cost method for "
                f"{blk.unit_model._tech_type}. Specify one of the following "
                f"cost methods in the unit's YAML file: {valid_methods}"
            )

    @staticmethod
    def cost_HRCS_clarifier(blk):
        """
        Method for costing a clarifier unit in a high-rate contact stabilization (HRCS) process.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        HRT, size_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["HRT", "sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * HRT * size_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )

        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.ferric_chloride_demand[t0], "ferric_chloride"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import pyomo.environ as pyo
import idaes.core.util.scaling as iscale
import idaes.core.util.model_statistics as istat

from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent
from idaes.core import (
    FlowsheetBlock,
    EnergyBalanceType,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelCostingBlock,
)
from idaes.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester
from idaes.core.util.exceptions import ConfigurationError
from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    DiffusivityCalculation,
)
from watertap.unit_models.gac import (
    GAC,
    FilmTransferCoefficientType,
    SurfaceDiffusionCoefficientType,
)
from watertap.costing import WaterTAPCosting

__author__ = "Hunter Barber"

solver = get_solver()

# -----------------------------------------------------------------------------
class TestGACSimplified:
    @pytest.fixture(scope="class")
    def gac_frame_simplified(self):
        ms = pyo.ConcreteModel()
        ms.fs = FlowsheetBlock(dynamic=False)

        ms.fs.properties = MCASParameterBlock(
            solute_list=["DCE"],
            mw_data={"H2O": 0.018, "DCE": 0.09896},
            ignore_neutral_charge=True,
        )
        ms.fs.unit = GAC(
            property_package=ms.fs.properties,
            film_transfer_coefficient_type="fixed",
            surface_diffusion_coefficient_type="fixed",
        )

        # feed specifications
        unit_feed = ms.fs.unit.process_flow.properties_in[0]
        unit_feed.pressure.fix(101325)
        unit_feed.temperature.fix(273.15 + 25)
        unit_feed.flow_mol_phase_comp["Liq", "H2O"].fix(55555.55426666667)
        unit_feed.flow_mol_phase_comp["Liq", "DCE"].fix(0.0002344381568310428)

        # trial problem from Hand, 1984 for removal of trace DCE
        # adsorption isotherm
        ms.fs.unit.freund_k.fix(37.9e-6 * (1e6**0.8316))
        ms.fs.unit.freund_ninv.fix(0.8316)
        # gac particle specifications
        ms.fs.unit.particle_dens_app.fix(722)
        ms.fs.unit.particle_dia.fix(0.00106)
        # adsorber bed specifications
        ms.fs.unit.ebct.fix(300)  # seconds
        ms.fs.unit.bed_voidage.fix(0.449)
        ms.fs.unit.bed_length.fix(6)  # assumed
        # design spec
        ms.fs.unit.conc_ratio_replace.fix(0.50)
        # parameters
        ms.fs.unit.kf.fix(3.29e-5)
        ms.fs.unit.ds.fix(1.77e-13)
        ms.fs.unit.a0.fix(3.68421)
        ms.fs.unit.a1.fix(13.1579)
        ms.fs.unit.b0.fix(0.784576)
        ms.fs.unit.b1.fix(0.239663)
        ms.fs.unit.b2.fix(0.484422)
        ms.fs.unit.b3.fix(0.003206)
        ms.fs.unit.b4.fix(0.134987)

        return ms

    @pytest.mark.unit
    def test_simplified_config(self, gac_frame_simplified):
        ms = gac_frame_simplified
        u_config = ms.fs.unit.config

        # check unit config arguments
        assert len(u_config) == 12

        assert not u_config.dynamic
        assert not u_config.has_holdup
        assert u_config.material_balance_type == MaterialBalanceType.useDefault
        assert u_config.energy_balance_type == EnergyBalanceType.none
        assert u_config.momentum_balance_type == MomentumBalanceType.pressureTotal
        assert u_config.finite_elements_ss_approximation == 5
        assert (
            u_config.film_transfer_coefficient_type == FilmTransferCoefficientType.fixed
        )
        assert (
            u_config.surface_diffusion_coefficient_type
            == SurfaceDiffusionCoefficientType.fixed
        )

        # check properties
        assert u_config.property_package is ms.fs.properties
        assert len(u_config.property_package.solute_set) == 1
        assert len(u_config.property_package.solvent_set) == 1
        assert ms.fs.properties.config.diffus_calculation == DiffusivityCalculation.none

    @pytest.mark.unit
    def test_simplified_build(self, gac_frame_simplified):
        ms = gac_frame_simplified

        # test units
        assert assert_units_consistent(ms) is None

        # test ports
        port_lst = ["inlet", "outlet", "adsorbed"]
        for port_str in port_lst:
            port = getattr(ms.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)

        # test statistics
        assert istat.number_variables(ms) == 100
        assert istat.number_total_constraints(ms) == 64
        assert istat.number_unused_variables(ms) == 11

    @pytest.mark.unit
    def test_simplified_dof(self, gac_frame_simplified):
        ms = gac_frame_simplified
        assert istat.degrees_of_freedom(ms) == 0

    @pytest.mark.unit
    def test_simplified_calculate_scaling(self, gac_frame_simplified):
        ms = gac_frame_simplified

        prop = ms.fs.properties
        prop.set_default_scaling("flow_mol_phase_comp", 1e-4, index=("Liq", "H2O"))
        prop.set_default_scaling("flow_mol_phase_comp", 1e4, index=("Liq", "DCE"))
        iscale.calculate_scaling_factors(ms)

        # check that all variables have scaling factors
        assert len(list(iscale.unscaled_variables_generator(ms))) == 0

    @pytest.mark.component
    def test_simplified_initialize(self, gac_frame_simplified):
        initialization_tester(gac_frame_simplified)

    @pytest.mark.component
    def test_simplified_var_scaling_init(self, gac_frame_simplified):
        ms = gac_frame_simplified
        assert len(list(iscale.badly_scaled_var_generator(ms, zero=1e-8))) == 0

    @pytest.mark.component
    def test_simplified_solve(self, gac_frame_simplified):
        ms = gac_frame_simplified
        results = solver.solve(ms)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)

    @pytest.mark.component
    def test_simplified_var_scaling_solve(self, gac_frame_simplified):
        ms = gac_frame_simplified
        assert len(list(iscale.badly_scaled_var_generator(ms, zero=1e-8))) == 0

    @pytest.mark.component
    def test_simplified_solution(self, gac_frame_simplified):
        ms = gac_frame_simplified
        gac = ms.fs.unit

        # Approx data pulled from graph in Hand, 1984 at ~30 days
        # 30 days adjusted to actual solution to account for web plot data extraction error within reason
        # values calculated by hand and match those reported in Hand, 1984
        assert pytest.approx(0.0005178, rel=1e-3) == pyo.value(gac.equil_conc)
        assert pytest.approx(19780, rel=1e-3) == pyo.value(gac.dg)
        assert pytest.approx(6.113, rel=1e-3) == pyo.value(gac.N_Bi)
        assert pytest.approx(35.68, rel=1e-3) == pyo.value(gac.min_N_St)
        assert pytest.approx(0.9882, rel=1e-3) == pyo.value(gac.throughput)
        assert pytest.approx(468.4, rel=1e-3) == pyo.value(gac.min_residence_time)
        assert pytest.approx(134.7, rel=1e-3) == pyo.value(gac.residence_time)
        assert pytest.approx(9153000, rel=1e-3) == pyo.value(gac.min_operational_time)
        assert pytest.approx(2554000, rel=1e-3) == pyo.value(gac.operational_time)
        assert pytest.approx(8514, rel=1e-3) == pyo.value(gac.bed_volumes_treated)


# -----------------------------------------------------------------------------
class TestGACRobust:
    @pytest.fixture(scope="class")
    def gac_frame_robust(self):
        mr = pyo.ConcreteModel()
        mr.fs = FlowsheetBlock(dynamic=False)

        mr.fs.properties = MCASParameterBlock(
            solute_list=["TCE"],
            mw_data={"H2O": 0.018, "TCE": 0.1314},
            diffus_calculation=DiffusivityCalculation.HaydukLaudie,
            molar_volume_data={("Liq", "TCE"): 9.81e-5},
            ignore_neutral_charge=True,
        )
        mr.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
        mr.fs.properties.dens_mass_const = 999.7
        mr.fs.unit = GAC(
            property_package=mr.fs.properties,
            film_transfer_coefficient_type="fixed",
            surface_diffusion_coefficient_type="fixed",
            finite_elements_ss_approximation=9,
        )

        # feed specifications
        unit_feed = mr.fs.unit.process_flow.properties_in[0]
        unit_feed.pressure.fix(101325)
        unit_feed.temperature.fix(273.15 + 25)
        unit_feed.flow_mol_phase_comp["Liq", "H2O"].fix(823.8)
        unit_feed.flow_mol_phase_comp["Liq", "TCE"].fix(5.6444e-05)

        # trial problem from Crittenden, 2012 for removal of TCE
        # adsorption isotherm
        mr.fs.unit.freund_k.fix(1062e-6 * (1e6**0.48))
        mr.fs.unit.freund_ninv.fix(0.48)
        # gac particle specifications
        mr.fs.unit.particle_dens_app.fix(803.4)
        mr.fs.unit.particle_dia.fix(0.001026)
        # adsorber bed specifications
        mr.fs.unit.ebct.fix(10 * 60)
        mr.fs.unit.bed_voidage.fix(0.44)
        mr.fs.unit.velocity_sup.fix(5 / 3600)
        # design spec
        mr.fs.unit.conc_ratio_replace.fix(0.80)
        # parameters
        mr.fs.unit.ds.fix(1.24e-14)
        mr.fs.unit.kf.fix(3.73e-05)
        mr.fs.unit.a0.fix(0.8)
        mr.fs.unit.a1.fix(0)
        mr.fs.unit.b0.fix(0.023)
        mr.fs.unit.b1.fix(0.793673)
        mr.fs.unit.b2.fix(0.039324)
        mr.fs.unit.b3.fix(0.009326)
        mr.fs.unit.b4.fix(0.08275)

        return mr

    @pytest.mark.unit
    def test_robust_config(self, gac_frame_robust):
        mr = gac_frame_robust
        u_config = mr.fs.unit.config

        # check unit config arguments
        assert len(u_config) == 12

        assert not u_config.dynamic
        assert not u_config.has_holdup
        assert u_config.material_balance_type == MaterialBalanceType.useDefault
        assert u_config.energy_balance_type == EnergyBalanceType.none
        assert u_config.momentum_balance_type == MomentumBalanceType.pressureTotal
        assert u_config.finite_elements_ss_approximation == 9
        assert (
            u_config.film_transfer_coefficient_type == FilmTransferCoefficientType.fixed
        )
        assert (
            u_config.surface_diffusion_coefficient_type
            == SurfaceDiffusionCoefficientType.fixed
        )

        # check properties
        assert u_config.property_package is mr.fs.properties
        assert len(u_config.property_package.solute_set) == 1
        assert len(u_config.property_package.solvent_set) == 1
        assert (
            mr.fs.properties.config.diffus_calculation
            == DiffusivityCalculation.HaydukLaudie
        )

    @pytest.mark.unit
    def test_robust_build(self, gac_frame_robust):
        mr = gac_frame_robust

        # test units
        assert assert_units_consistent(mr) is None

        # test ports
        port_lst = ["inlet", "outlet", "adsorbed"]
        for port_str in port_lst:
            port = getattr(mr.fs.unit, port_str)
            assert len(port.vars) == 3
            assert isinstance(port, Port)

        # test statistics
        assert istat.number_variables(mr) == 119
        assert istat.number_total_constraints(mr) == 84
        assert istat.number_unused_variables(mr) == 10

    @pytest.mark.unit
    def test_robust_dof(self, gac_frame_robust):
        mr = gac_frame_robust
        assert istat.degrees_of_freedom(mr) == 0

    @pytest.mark.unit
    def test_robust_calculate_scaling(self, gac_frame_robust):
        mr = gac_frame_robust

        prop = mr.fs.properties
        prop.set_default_scaling("flow_mol_phase_comp", 1e-2, index=("Liq", "H2O"))
        prop.set_default_scaling("flow_mol_phase_comp", 1e5, index=("Liq", "TCE"))
        iscale.calculate_scaling_factors(mr)

        # check that all variables have scaling factors
        assert len(list(iscale.unscaled_variables_generator(mr))) == 0

    @pytest.mark.component
    def test_robust_initialize(self, gac_frame_robust):
        initialization_tester(gac_frame_robust)

    @pytest.mark.component
    def test_robust_var_scaling_init(self, gac_frame_robust):
        mr = gac_frame_robust
        assert len(list(iscale.badly_scaled_var_generator(mr, zero=1e-8))) == 0

    @pytest.mark.component
    def test_robust_solve(self, gac_frame_robust):
        mr = gac_frame_robust
        results = solver.solve(mr)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)

    @pytest.mark.component
    def test_robust_var_scaling_solve(self, gac_frame_robust):
        mr = gac_frame_robust
        assert len(list(iscale.badly_scaled_var_generator(mr, zero=1e-8))) == 0

    @pytest.mark.component
    def test_robust_solution(self, gac_frame_robust):
        mr = gac_frame_robust
        gac = mr.fs.unit

        # values calculated by hand and match those reported in Crittenden, 2012
        assert pytest.approx(0.02097, rel=1e-3) == pyo.value(gac.equil_conc)
        assert pytest.approx(42890, rel=1e-3) == pyo.value(gac.dg)
        assert pytest.approx(45.79, rel=1e-3) == pyo.value(gac.N_Bi)
        assert pytest.approx(36.64, rel=1e-3) == pyo.value(gac.min_N_St)
        assert pytest.approx(1.139, rel=1e-3) == pyo.value(gac.throughput)
        assert pytest.approx(395.9, rel=1e-3) == pyo.value(gac.min_residence_time)
        assert pytest.approx(264.0, rel=1e-3) == pyo.value(gac.residence_time)
        assert pytest.approx(19340000, rel=1e-3) == pyo.value(gac.min_operational_time)
        assert pytest.approx(13690000, rel=1e-3) == pyo.value(gac.operational_time)
        assert pytest.approx(22810, rel=1e-3) == pyo.value(gac.bed_volumes_treated)
        assert pytest.approx(0.003157, rel=1e-3) == pyo.value(gac.velocity_int)
        assert pytest.approx(0.8333, rel=1e-3) == pyo.value(gac.bed_length)
        assert pytest.approx(10.68, rel=1e-3) == pyo.value(gac.bed_area)
        assert pytest.approx(8.900, rel=1e-3) == pyo.value(gac.bed_volume)
        assert pytest.approx(3.688, rel=1e-3) == pyo.value(gac.bed_diameter)
        assert pytest.approx(4004, rel=1e-3) == pyo.value(gac.bed_mass_gac)
        assert pytest.approx(0.2287, rel=1e-3) == pyo.value(gac.conc_ratio_avg)
        assert pytest.approx(6462000, rel=1e-3) == pyo.value(
            gac.ele_operational_time[1]
        )

    @pytest.mark.component
    def test_robust_reporting(self, gac_frame_robust):
        mr = gac_frame_robust
        mr.fs.unit.report()

    @pytest.mark.component
    def test_robust_costing_pressure(self, gac_frame_robust):
        mr = gac_frame_robust

        mr.fs.costing = WaterTAPCosting()
        mr.fs.costing.base_currency = pyo.units.USD_2020

        mr.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=mr.fs.costing
        )
        mr.fs.costing.cost_process()

        # testing gac costing block dof and initialization
        assert assert_units_consistent(mr) is None
        assert istat.degrees_of_freedom(mr) == 0
        mr.fs.unit.costing.initialize()

        # solve
        results = solver.solve(mr)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)

        cost = mr.fs.unit.costing
        # Check for known cost solution of default twin alternating contactors
        assert pyo.value(mr.fs.costing.gac_pressure.num_contactors_op) == 1
        assert pyo.value(mr.fs.costing.gac_pressure.num_contactors_redundant) == 1
        assert pytest.approx(56900, rel=1e-3) == pyo.value(cost.contactor_cost)
        assert pytest.approx(4.359, rel=1e-3) == pyo.value(cost.adsorbent_unit_cost)
        assert pytest.approx(17450, rel=1e-3) == pyo.value(cost.adsorbent_cost)
        assert pytest.approx(81690, rel=1e-3) == pyo.value(cost.other_process_cost)
        assert pytest.approx(2.0 * 156000, rel=1e-3) == pyo.value(cost.capital_cost)
        assert pytest.approx(12680, rel=1e-3) == pyo.value(cost.gac_makeup_cost)
        assert pytest.approx(27660, rel=1e-3) == pyo.value(cost.gac_regen_cost)
        assert pytest.approx(0.01631, rel=1e-3) == pyo.value(cost.energy_consumption)
        assert pytest.approx(40370, rel=1e-3) == pyo.value(cost.fixed_operating_cost)

    @pytest.mark.component
    def test_robust_costing_gravity(self, gac_frame_robust):
        mr_grav = gac_frame_robust

        mr_grav.fs.costing = WaterTAPCosting()
        mr_grav.fs.costing.base_currency = pyo.units.USD_2020

        mr_grav.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=mr_grav.fs.costing,
            costing_method_arguments={"contactor_type": "gravity"},
        )
        mr_grav.fs.costing.cost_process()

        # testing gac costing block dof and initialization
        assert assert_units_consistent(mr_grav) is None
        assert istat.degrees_of_freedom(mr_grav) == 0
        mr_grav.fs.unit.costing.initialize()

        # solve
        results = solver.solve(mr_grav)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)

        cost = mr_grav.fs.unit.costing
        # Check for known cost solution of default twin alternating contactors
        assert pyo.value(mr_grav.fs.costing.gac_gravity.num_contactors_op) == 1
        assert pyo.value(mr_grav.fs.costing.gac_gravity.num_contactors_redundant) == 1
        assert pytest.approx(163200, rel=1e-3) == pyo.value(cost.contactor_cost)
        assert pytest.approx(4.359, rel=1e-3) == pyo.value(cost.adsorbent_unit_cost)
        assert pytest.approx(17450, rel=1e-3) == pyo.value(cost.adsorbent_cost)
        assert pytest.approx(159500, rel=1e-3) == pyo.value(cost.other_process_cost)
        assert pytest.approx(2.0 * 340200, rel=1e-3) == pyo.value(cost.capital_cost)
        assert pytest.approx(12680, rel=1e-3) == pyo.value(cost.gac_makeup_cost)
        assert pytest.approx(27660, rel=1e-3) == pyo.value(cost.gac_regen_cost)
        assert pytest.approx(2.476, rel=1e-3) == pyo.value(cost.energy_consumption)
        assert pytest.approx(40370, rel=1e-3) == pyo.value(cost.fixed_operating_cost)

    @pytest.mark.component
    def test_robust_costing_modular_contactors(self, gac_frame_robust):
        mr = gac_frame_robust

        mr.fs.costing = WaterTAPCosting()
        mr.fs.costing.base_currency = pyo.units.USD_2020

        mr.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=mr.fs.costing
        )
        mr.fs.costing.cost_process()

        mr.fs.costing.gac_pressure.num_contactors_op.fix(4)
        mr.fs.costing.gac_pressure.num_contactors_redundant.fix(2)

        results = solver.solve(mr)

        cost = mr.fs.unit.costing
        # Check for known cost solution when changing volume scale of vessels in parallel
        assert pyo.value(mr.fs.costing.gac_pressure.num_contactors_op) == 4
        assert pyo.value(mr.fs.costing.gac_pressure.num_contactors_redundant) == 2
        assert pytest.approx(89040, rel=1e-3) == pyo.value(cost.contactor_cost)
        assert pytest.approx(69690, rel=1e-3) == pyo.value(cost.other_process_cost)
        assert pytest.approx(2.0 * 176200, rel=1e-3) == pyo.value(cost.capital_cost)

    @pytest.mark.component
    def test_robust_costing_max_gac_ref(self, gac_frame_robust):
        mr = gac_frame_robust

        # scale flow up 10x
        unit_feed = mr.fs.unit.process_flow.properties_in[0]
        unit_feed.flow_mol_phase_comp["Liq", "H2O"].fix(10 * 824.0736620370348)
        unit_feed.flow_mol_phase_comp["Liq", "TCE"].fix(10 * 5.644342973110135e-05)

        mr.fs.costing = WaterTAPCosting()
        mr.fs.costing.base_currency = pyo.units.USD_2020

        mr.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=mr.fs.costing
        )
        mr.fs.costing.cost_process()
        # not necessarily an optimum solution because poor scaling
        # but just checking the conditional
        results = solver.solve(mr)

        # Check for bed_mass_gac_cost_ref to be overwritten
        # if bed_mass_gac is greater than bed_mass_gac_cost_max_ref
        assert pyo.value(mr.fs.unit.bed_mass_gac) > pyo.value(
            mr.fs.costing.gac_pressure.bed_mass_max_ref
        )
        assert pyo.value(mr.fs.unit.costing.bed_mass_gac_ref) == (
            pytest.approx(pyo.value(mr.fs.costing.gac_pressure.bed_mass_max_ref), 1e-5)
        )


# -----------------------------------------------------------------------------
class TestGACMulti:
    @pytest.fixture(scope="class")
    def gac_frame_multi(self):
        mm = pyo.ConcreteModel()
        mm.fs = FlowsheetBlock(dynamic=False)

        # inserting arbitrary BackGround Solutes, Cations, and Anions to check handling
        # arbitrary diffusivity data for non-target species
        mm.fs.properties = MCASParameterBlock(
            solute_list=["TCE", "BGSOL", "BGCAT", "BGAN"],
            mw_data={
                "H2O": 0.018,
                "TCE": 0.1314,
                "BGSOL": 0.1,
                "BGCAT": 0.1,
                "BGAN": 0.1,
            },
            charge={"BGCAT": 1, "BGAN": -2},
            diffus_calculation=DiffusivityCalculation.HaydukLaudie,
            molar_volume_data={("Liq", "TCE"): 9.81e-5},
            diffusivity_data={
                ("Liq", "BGSOL"): 1e-5,
                ("Liq", "BGCAT"): 1e-5,
                ("Liq", "BGAN"): 1e-5,
            },
            ignore_neutral_charge=True,
        )
        mm.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
        mm.fs.properties.dens_mass_const = 1000
        # testing target_species arg
        mm.fs.unit = GAC(
            property_package=mm.fs.properties,
            film_transfer_coefficient_type="calculated",
            surface_diffusion_coefficient_type="calculated",
            target_species={"TCE"},
        )

        unit_feed = mm.fs.unit.process_flow.properties_in[0]
        # feed specifications
        unit_feed.pressure.fix(101325)  # feed pressure [Pa]
        unit_feed.temperature.fix(273.15 + 25)  # feed temperature [K]
        unit_feed.flow_mol_phase_comp["Liq", "H2O"].fix(824.0736620370348)
        unit_feed.flow_mol_phase_comp["Liq", "TCE"].fix(5.644342973110135e-05)
        unit_feed.flow_mol_phase_comp["Liq", "BGSOL"].fix(5e-05)
        unit_feed.flow_mol_phase_comp["Liq", "BGCAT"].fix(2e-05)
        unit_feed.flow_mol_phase_comp["Liq", "BGAN"].fix(1e-05)

        # trial problem from Crittenden, 2012 for removal of TCE
        # adsorption isotherm
        mm.fs.unit.freund_k.fix(1062e-6 * (1e6**0.48))
        mm.fs.unit.freund_ninv.fix(0.48)
        # gac particle specifications
        mm.fs.unit.particle_dens_app.fix(803.4)
        mm.fs.unit.particle_dia.fix(0.001026)
        # adsorber bed specifications
        mm.fs.unit.ebct.fix(10 * 60)
        mm.fs.unit.bed_voidage.fix(0.44)
        mm.fs.unit.velocity_sup.fix(5 / 3600)
        # design spec
        mm.fs.unit.conc_ratio_replace.fix(0.80)
        # parameters
        mm.fs.unit.particle_porosity.fix(0.641)
        mm.fs.unit.tort.fix(1)
        mm.fs.unit.spdfr.fix(1)
        mm.fs.unit.shape_correction_factor.fix(1.5)
        mm.fs.unit.a0.fix(0.8)
        mm.fs.unit.a1.fix(0)
        mm.fs.unit.b0.fix(0.023)
        mm.fs.unit.b1.fix(0.793673)
        mm.fs.unit.b2.fix(0.039324)
        mm.fs.unit.b3.fix(0.009326)
        mm.fs.unit.b4.fix(0.08275)

        return mm

    @pytest.mark.unit
    def test_multi_config(self, gac_frame_multi):
        mm = gac_frame_multi
        u_config = mm.fs.unit.config

        # checking non-unity solute set and nonzero ion set handling
        assert len(u_config.property_package.solute_set) == 4
        assert len(u_config.property_package.solvent_set) == 1
        assert len(u_config.property_package.ion_set) == 2
        assert (
            u_config.film_transfer_coefficient_type
            == FilmTransferCoefficientType.calculated
        )
        assert (
            u_config.surface_diffusion_coefficient_type
            == SurfaceDiffusionCoefficientType.calculated
        )
        assert (
            mm.fs.properties.config.diffus_calculation
            == DiffusivityCalculation.HaydukLaudie
        )

        assert istat.degrees_of_freedom(mm) == 0

    @pytest.mark.unit
    def test_multi_calculate_scaling(self, gac_frame_multi):
        mm = gac_frame_multi

        prop = mm.fs.properties
        prop.set_default_scaling("flow_mol_phase_comp", 1e-2, index=("Liq", "H2O"))
        for j in prop.ion_set | prop.solute_set:
            prop.set_default_scaling("flow_mol_phase_comp", 1e5, index=("Liq", j))

        iscale.calculate_scaling_factors(mm)
        initialization_tester(gac_frame_multi)

        # check that all variables have scaling factors
        assert len(list(iscale.unscaled_variables_generator(mm))) == 0

    @pytest.mark.unit
    def test_multi_var_scaling_init(self, gac_frame_multi):
        mm = gac_frame_multi
        assert len(list(iscale.badly_scaled_var_generator(mm, zero=1e-8))) == 0

    @pytest.mark.component
    def test_multi_solve(self, gac_frame_multi):
        mm = gac_frame_multi
        results = solver.solve(mm)

        # Check for optimal solution
        assert pyo.check_optimal_termination(results)

    @pytest.mark.unit
    def test_multi_var_scaling_solve(self, gac_frame_multi):
        mm = gac_frame_multi
        assert len(list(iscale.badly_scaled_var_generator(mm, zero=1e-8))) == 0

    @pytest.mark.component
    def test_multi_solution(self, gac_frame_multi):
        mm = gac_frame_multi

        # only checking for variables new to configuration options
        assert pytest.approx(2.473, rel=1e-3) == pyo.value(mm.fs.unit.N_Re)
        assert pytest.approx(2001, rel=1e-3) == pyo.value(mm.fs.unit.N_Sc)
        assert pytest.approx(2.600e-5, rel=1e-3) == pyo.value(mm.fs.unit.kf)
        assert pytest.approx(1.245e-14, rel=1e-3) == pyo.value(mm.fs.unit.ds)

    @pytest.mark.component
    def test_multi_reporting(self, gac_frame_multi):
        mm = gac_frame_multi
        mm.fs.unit.report()


# -----------------------------------------------------------------------------
class TestGACErrorLog:
    @pytest.mark.unit
    def test_error(self):

        with pytest.raises(
            ConfigurationError,
            match="'target species' is not specified for the GAC unit model, "
            "either specify 'target species' argument or reduce solute set "
            "to a single component",
        ):
            me = pyo.ConcreteModel()
            me.fs = FlowsheetBlock(dynamic=False)

            # inserting arbitrary BackGround Solutes, Cations, and Anions to check handling
            # arbitrary diffusivity data for non-target species
            me.fs.properties = MCASParameterBlock(
                solute_list=["TCE", "BGSOL", "BGCAT", "BGAN"],
                mw_data={
                    "H2O": 0.018,
                    "TCE": 0.1314,
                    "BGSOL": 0.1,
                    "BGCAT": 0.1,
                    "BGAN": 0.1,
                },
                charge={"BGCAT": 1, "BGAN": -2},
                diffus_calculation=DiffusivityCalculation.HaydukLaudie,
                molar_volume_data={("Liq", "TCE"): 9.81e-5},
                diffusivity_data={
                    ("Liq", "BGSOL"): 1e-5,
                    ("Liq", "BGCAT"): 1e-5,
                    ("Liq", "BGAN"): 1e-5,
                },
                ignore_neutral_charge=True,
            )
            me.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
            me.fs.properties.dens_mass_const = 1000

            # testing target_species arg
            me.fs.unit = GAC(
                property_package=me.fs.properties,
                film_transfer_coefficient_type="calculated",
                surface_diffusion_coefficient_type="calculated",
            )

        with pytest.raises(
            ConfigurationError,
            match="fs.unit received invalid argument for contactor_type:"
            " vessel. Argument must be a member of the ContactorType Enum.",
        ):
            me = pyo.ConcreteModel()
            me.fs = FlowsheetBlock(dynamic=False)

            me.fs.properties = MCASParameterBlock(
                solute_list=["TCE"],
                mw_data={"H2O": 0.018, "TCE": 0.1314},
                diffus_calculation=DiffusivityCalculation.HaydukLaudie,
                molar_volume_data={("Liq", "TCE"): 9.81e-5},
                charge={"TCE": 0},
            )
            me.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
            me.fs.properties.dens_mass_const = 1000

            me.fs.unit = GAC(
                property_package=me.fs.properties,
                film_transfer_coefficient_type="calculated",
                surface_diffusion_coefficient_type="calculated",
            )

            me.fs.costing = WaterTAPCosting()
            me.fs.costing.base_currency = pyo.units.USD_2020

            me.fs.unit.costing = UnitModelCostingBlock(
                flowsheet_costing_block=me.fs.costing,
                costing_method_arguments={"contactor_type": "vessel"},
            )

        with pytest.raises(
            ConfigurationError,
            match="item 0 within 'target_species' list is not of data type str",
        ):
            me = pyo.ConcreteModel()
            me.fs = FlowsheetBlock(dynamic=False)

            # inserting arbitrary BackGround Solutes, Cations, and Anions to check handling
            # arbitrary diffusivity data for non-target species
            me.fs.properties = MCASParameterBlock(
                solute_list=["TCE", "BGSOL", "BGCAT", "BGAN"],
                mw_data={
                    "H2O": 0.018,
                    "TCE": 0.1314,
                    "BGSOL": 0.1,
                    "BGCAT": 0.1,
                    "BGAN": 0.1,
                },
                charge={"BGCAT": 1, "BGAN": -2},
                diffus_calculation=DiffusivityCalculation.HaydukLaudie,
                molar_volume_data={("Liq", "TCE"): 9.81e-5},
                diffusivity_data={
                    ("Liq", "BGSOL"): 1e-5,
                    ("Liq", "BGCAT"): 1e-5,
                    ("Liq", "BGAN"): 1e-5,
                },
                ignore_neutral_charge=True,
            )
            me.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
            me.fs.properties.dens_mass_const = 1000

            # testing target_species arg
            me.fs.unit = GAC(
                property_package=me.fs.properties,
                film_transfer_coefficient_type="calculated",
                surface_diffusion_coefficient_type="calculated",
                target_species=range(2),
            )

        with pytest.raises(
            ConfigurationError,
            match="item species within 'target_species' list is not in 'component_list",
        ):
            me = pyo.ConcreteModel()
            me.fs = FlowsheetBlock(dynamic=False)

            # inserting arbitrary BackGround Solutes, Cations, and Anions to check handling
            # arbitrary diffusivity data for non-target species
            me.fs.properties = MCASParameterBlock(
                solute_list=["TCE", "BGSOL", "BGCAT", "BGAN"],
                mw_data={
                    "H2O": 0.018,
                    "TCE": 0.1314,
                    "BGSOL": 0.1,
                    "BGCAT": 0.1,
                    "BGAN": 0.1,
                },
                charge={"BGCAT": 1, "BGAN": -2},
                diffus_calculation=DiffusivityCalculation.HaydukLaudie,
                molar_volume_data={("Liq", "TCE"): 9.81e-5},
                diffusivity_data={
                    ("Liq", "BGSOL"): 1e-5,
                    ("Liq", "BGCAT"): 1e-5,
                    ("Liq", "BGAN"): 1e-5,
                },
                ignore_neutral_charge=True,
            )
            me.fs.properties.visc_d_phase["Liq"] = 1.3097e-3
            me.fs.properties.dens_mass_const = 1000

            # testing target_species arg
            me.fs.unit = GAC(
                property_package=me.fs.properties,
                film_transfer_coefficient_type="calculated",
                surface_diffusion_coefficient_type="calculated",
                target_species={"species": "TCE"},
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a Magprex reactor unit
for struvite precipitation.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Marcus Holly"


@declare_process_block_class("MagprexZO")
class MagprexZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Magprex reactor unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "magprex"

        build_sido_reactive(self)
        constant_intensity(self)

        self.magnesium_chloride_dosage = Var(
            units=pyunits.dimensionless,
            bounds=(0, None),
            doc="Dosage of magnesium chloride per phosphates",
        )

        self._fixed_perf_vars.append(self.magnesium_chloride_dosage)

        self._perf_var_dict[
            "Dosage of magnesium chloride per phosphates"
        ] = self.magnesium_chloride_dosage

        self.MgCl2_flowrate = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.hr,
            bounds=(0, None),
            doc="Magnesium chloride flowrate",
        )

        self._perf_var_dict["Magnesium Chloride Demand"] = self.MgCl2_flowrate

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for magnesium chloride demand based on sludge flowrate.",
        )
        def MgCl2_demand(b, t):
            return b.MgCl2_flowrate[t] == (
                b.magnesium_chloride_dosage
                * pyunits.convert(
                    b.properties_in[t].flow_mass_comp["phosphates"],
                    to_units=pyunits.kg / pyunits.hour,
                )
            )

    @property
    def default_costing_method(self):
        return self.cost_magprex

    @staticmethod
    def cost_magprex(blk):
        """
        Method for costing Magprex reactor unit.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        HRT, size_cost = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["HRT", "sizing_cost"],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol * HRT * size_cost,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.MgCl2_flowrate[t0], "magnesium_chloride"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""Zero order nanofiltration model based on specifying solvent flux and solute rejection"""

from pyomo.environ import ConcreteModel, Constraint
from idaes.core import FlowsheetBlock
from watertap.unit_models.nanofiltration_ZO import NanofiltrationZO
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    constraint_scaling_transform,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)


def build_ZONF(m, base="ion"):
    """
    Builds a ZONF model based on specified rejection and solvent flux.
    Requires prop_ion property package.
    """

    if base not in ["ion"]:
        raise ValueError(
            "Unexpected property base {base} for build_ZONF" "".format(base=base)
        )
    prop = property_models.get_prop(m, base=base)

    m.fs.NF = NanofiltrationZO(property_package=prop, has_pressure_change=False)

    # specify
    m.fs.NF.flux_vol_solvent.fix(1.67e-6)
    m.fs.NF.area.fix(500)
    m.fs.NF.properties_permeate[0].pressure.fix(101325)

    m.fs.NF.rejection_phase_comp[0, "Liq", "Na"].fix(0.01)
    m.fs.NF.rejection_phase_comp[0, "Liq", "Ca"].fix(0.79)
    m.fs.NF.rejection_phase_comp[0, "Liq", "Mg"].fix(0.94)
    m.fs.NF.rejection_phase_comp[0, "Liq", "SO4"].fix(0.87)
    m.fs.NF.rejection_phase_comp[
        0, "Liq", "Cl"
    ] = 0.15  # guess, but electroneutrality enforced below
    charge_comp = {"Na": 1, "Ca": 2, "Mg": 2, "SO4": -2, "Cl": -1}
    m.fs.NF.eq_electroneutrality = Constraint(
        expr=0
        == sum(
            charge_comp[j]
            * m.fs.NF.feed_side.properties_out[0].conc_mol_phase_comp["Liq", j]
            for j in charge_comp
        )
    )
    constraint_scaling_transform(m.fs.NF.eq_electroneutrality, 1)


def solve_ZONF(base="ion"):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    property_models.build_prop(m, base=base)
    build_ZONF(m, base=base)
    property_models.specify_feed(m.fs.NF.feed_side.properties_in[0], base="ion")

    check_dof(m)
    calculate_scaling_factors(m)
    solve_block(m)

    m.fs.NF.inlet.display()
    m.fs.NF.permeate.display()
    m.fs.NF.retentate.display()

    return m


if __name__ == "__main__":
    solve_ZONF(base="ion")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests that documentation exists for all zero-order unit models.
To create new documentation, update the unit classification excel sheet and run the automation script.
"""
import pytest
import os

from pyomo.common.fileutils import this_file_dir
from pathlib import Path

this_file_dir = this_file_dir()

doc_path = os.path.join(
    this_file_dir,
    "..",
    "..",
    "..",
    "..",
    "docs",
)

doc_path = Path(doc_path).resolve()

rst_path = os.path.join(
    doc_path,
    "technical_reference",
    "unit_models",
    "zero_order_unit_models",
)

py_path = os.path.join(this_file_dir, "..")


def generate_model_list(path, extension, exclude):
    file_list = []
    for file in os.listdir(path):
        filename = os.fsdecode(file)
        if filename.endswith(extension) and filename not in exclude:
            file_path = os.path.join(
                path,
                file,
            )
            assert os.path.isfile(file_path)
            file_list.append(filename[: -len(extension)])
    return file_list


@pytest.mark.integration
@pytest.mark.skipif(
    not os.path.exists(doc_path / "index.rst"), reason="No RST files found"
)
def test_lists_match():

    rst_list = generate_model_list(rst_path, ".rst", ("index.rst",))
    py_list = generate_model_list(py_path, ".py", ("__init__.py",))

    # remove the AOPMixin
    py_list.remove("aop_addition_zo")

    # Run pytest -vv to see how the lists are different
    assert sorted(py_list) == sorted(rst_list)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for fsapi module
"""
import logging
from pathlib import Path
import pytest
import tempfile

from pyomo.environ import units as pyunits
from pyomo.environ import Var, value
from pyomo.environ import SolverStatus, TerminationCondition

from watertap.examples.flowsheets.case_studies.seawater_RO_desalination import (
    seawater_RO_desalination as RO,
)
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.metab import (
    metab_ui as MU,
)

from watertap.ui import fsapi


_log = logging.getLogger("idaes.watertap.ui.fsapi")
_log.setLevel(logging.DEBUG)

ERD_TYPE = "pressure_exchanger"

# Fake status=OK solver result


class FAKE_FLOWSHEET:
    fs = "fs"
    trash = "true"


class SOLVE_RESULT_OK:
    class SOLVE_STATUS:
        status = SolverStatus.ok
        termination_condition = TerminationCondition.optimal

    solver = SOLVE_STATUS


def build_ro(build_options=None, **kwargs):
    model = RO.build_flowsheet(erd_type=ERD_TYPE)
    return model


def solve_ro(flowsheet=None):
    assert flowsheet
    return {"solved": True}


class InputCategory:
    """Names for input categories"""

    feed = "Feed"
    hydrogen = "Hydrogen reactor"
    methane = "Methane reactor"
    system = "System parameters"


class OutputCategory:
    """Names for output categories"""

    feed = "Feed"
    levelized_costs = "Levelized costs"
    normalized_costs = "Normalized costs"
    normalized_performance = "Normalized performance"
    product = "Product"
    capital_cost = "Capital cost"
    operating_cost = "Operating cost"
    revenue = "Revenue"


def export_to_ui(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    exports.add(
        obj=fs.feed.flow_vol[0],
        name="Flowrate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_rounding=1e-2,  # round to nearest 0.01
        description="Volumetric flowrate for the feed",
        is_input=True,
        input_category=InputCategory.feed,
        read_only=False,
        is_output=True,
        output_category=OutputCategory.feed,
    )


def flowsheet_interface(exports=True, solve_func=solve_ro):
    kwargs = {}
    if exports:
        kwargs["do_export"] = export_to_ui
    return fsapi.FlowsheetInterface(
        # leave out name and description to test auto-fill
        do_build=build_ro,
        do_solve=solve_func,
        **kwargs,
    )


def noop(*args, **kwargs):
    return


@pytest.mark.unit
def test_create_interface():
    with pytest.raises(ValueError):
        _ = flowsheet_interface(exports=False)
    fsi = flowsheet_interface()
    fs2 = fsapi.FlowsheetInterface(
        fs=fsi.fs_exp, do_build=noop, do_export=noop, do_solve=noop
    )
    assert fs2.fs_exp == fsi.fs_exp


@pytest.mark.unit
def test_build_noexport():
    with pytest.raises(ValueError):
        flowsheet_interface(exports=False)


@pytest.mark.unit
def test_build():
    fsi = flowsheet_interface()
    fsi.build(erd_type="pressure_exchanger")
    data = fsi.dict()
    print(data)
    assert "model_objects" in data
    assert len(data["model_objects"]) == 1


@pytest.mark.parametrize(
    "add_variant",
    [
        "obj_kwarg",
        "model_export_arg",
        "model_export_data_kwarg",
        "model_export_dict_data_kwarg",
    ],
)
@pytest.mark.unit
def test_actions(add_variant: str):
    fsi = flowsheet_interface()
    built = False
    # garbage = {"trash": True}
    garbage = FAKE_FLOWSHEET
    m = FAKE_FLOWSHEET
    v1 = Var(name="variable1")
    v1.construct()
    v1.value = 1
    print(v1.display())

    def fake_build(build_options=None, **kwargs):
        nonlocal built
        built = True
        nonlocal m
        m = build_ro()
        return m

    def fake_solve(flowsheet=None):
        # flowsheet passed in here should be what fake_build() returns
        assert flowsheet == m.fs
        return SOLVE_RESULT_OK

    def fake_export(flowsheet=None, exports=None, build_options=None, **kwargs):
        with pytest.raises(Exception):
            exports.add(obj=garbage)

        # NOTE we use exclusive variants here
        # to avoid triggering the error when adding an object with the same obj_key
        # which happens when multiple ModelExport are created from the same pyomo object
        if add_variant == "obj_kwarg":
            exports.add(obj=v1)  # form 1
        elif add_variant == "model_export_kwarg":
            ve1 = fsapi.ModelExport(obj=v1)
            exports.add(ve1)  # form 2
        elif add_variant == "model_export_data_kwarg":
            ve1 = fsapi.ModelExport(obj=v1)
            exports.add(data=ve1)  # form 3
        elif add_variant == "model_export_dict_data_kwarg":
            ve1 = fsapi.ModelExport(obj=v1)
            exports.add(data=ve1.dict())  # form 4
        with pytest.raises(ValueError):
            exports.add(v1, v1)

    fsi.add_action(fsapi.Actions.build, fake_build)
    fsi.add_action(fsapi.Actions.export, fake_export)
    fsi.add_action(fsapi.Actions.solve, fake_solve)
    fsi.build()
    fsi.solve()
    with pytest.raises(ValueError):
        fsi.run_action(fsapi.Actions.export)


class CSVTestSettings:
    """Settings for test_csv_exports used in other functions."""

    bad_obj = False
    bad_units = False


@pytest.mark.unit
def test_csv_exports():
    for i in range(3):
        if i == 1:
            CSVTestSettings.bad_obj, CSVTestSettings.bad_units = True, False
        elif i == 2:
            CSVTestSettings.bad_obj, CSVTestSettings.bad_units = False, True
        else:
            CSVTestSettings.bad_obj, CSVTestSettings.bad_units = False, False
        for export_func in (csv_from_tempfile, csv_from_localfile):
            fsi = fsapi.FlowsheetInterface(
                do_build=build_ro, do_solve=solve_ro, do_export=export_func
            )
            if i == 0:
                fsi.build()  # expect success
            else:
                # expect failure (bad_units or bad_obj)
                with pytest.raises(RuntimeError):
                    fsi.build()


def csv_from_tempfile(exports=None, flowsheet=None, **kwargs):
    with tempfile.TemporaryDirectory() as tempdir:
        f = Path(tempdir) / "fake.csv"
        populate_csv_exports(f.open("w"))
        exports.from_csv(file=f, flowsheet=flowsheet)


def csv_from_localfile(exports=None, flowsheet=None, **kwargs):
    path = Path(__file__).parent / "test.csv"
    populate_csv_exports(path.open("w"))
    try:
        exports.from_csv(file="test.csv", flowsheet=flowsheet)
    finally:
        path.unlink()


def populate_csv_exports(f):
    units = "units.foobar" if CSVTestSettings.bad_units else "units.m**3/units.s"
    obj = "dirt" if CSVTestSettings.bad_obj else "fs.feed.flow_vol[0]"
    rows = [
        "name,obj,description,ui_units,display_units,rounding,is_input,input_category,is_output,output_category",
        f"feed,{obj},feed flow volume,{units},m^3/s,3,TRUE,something,FALSE,",
    ]
    for row in rows:
        f.write(row)
        f.write("\n")


@pytest.mark.unit
def test_load():
    fsi = flowsheet_interface()
    fsi.build(erd_type="pressure_exchanger")
    # get some info
    var_key = list(fsi.fs_exp.model_objects.keys())[0]
    var_obj = fsi.fs_exp.model_objects[var_key].obj
    save_value = var_obj.value
    # serialize
    data = fsi.dict()
    # modify
    data["model_objects"][var_key]["value"] = -1000
    # reload
    fsi.load(data)
    # check
    assert fsi.fs_exp.model_objects[var_key].value == -1000

    # this time with a missing thing
    data = fsi.dict()
    # add another (fake) one
    data["model_objects"]["foobar"] = data["model_objects"][var_key].copy()
    # reload (fake one will be 'missing')
    try:
        fsi.load(data)
    except fsapi.FlowsheetInterface.MissingObjectError as err:
        for item in err.missing:
            print(f"Missing item: key={item.key}, name={item.name}")
        assert len(err.missing) == 1
        assert err.missing[0].key == "foobar"
    else:
        assert False, "Expected a MissingObjectError"


@pytest.mark.unit
def test_require_methods():
    fsi = flowsheet_interface()
    methods = ("do_export", "do_build", "do_solve")
    kwargs = {m: noop for m in methods}
    # make one method 'bad' at a time
    for meth in methods:
        # missing
        badkw = kwargs.copy()
        del badkw[meth]
        with pytest.raises(ValueError):
            _ = fsapi.FlowsheetInterface(fsi, **badkw)
        # not callable
        badkw = kwargs.copy()
        badkw[meth] = 1
        with pytest.raises(TypeError):
            _ = fsapi.FlowsheetInterface(fsi, **badkw)


@pytest.mark.component
def test_export_values():
    # get an interface
    fsi = flowsheet_interface()
    fsi.build()
    d1 = fsi.dict()

    # change one value
    key = list(fsi.fs_exp.model_objects.keys())[0]
    orig_value = value(fsi.fs_exp.model_objects[key].obj)
    new_value = orig_value + 1
    print(f"@@ orig_value = {orig_value}, new value = {new_value}")
    fsi.fs_exp.model_objects[key].obj.value = new_value

    # re-export
    fsi.export_values()
    d2 = fsi.dict()

    print("== original")
    print(d1)
    print("== modified")
    print(d2)

    # check that change happened
    assert d1 != d2


@pytest.mark.component
def test_export_values_build():
    # get an interface
    fsi = flowsheet_interface()
    d1 = fsi.dict()
    fsi.build()
    # after build, new values should be exported to fsi.fs_exp
    d2 = fsi.dict()
    assert d1 != d2


@pytest.mark.unit
def test_empty_solve():
    # try a fake solve
    fsi = flowsheet_interface()
    with pytest.raises(RuntimeError) as excinfo:
        fsi.build()
        fsi.solve()
    print(f"* RuntimeError: {excinfo.value}")


@pytest.mark.unit
def test_nonoptimal_termination():
    fsi = MU.export_to_ui()
    fsi.build()

    # pick a crazy value
    key = list(fsi.fs_exp.model_objects.keys())[0]
    orig_value = value(fsi.fs_exp.model_objects[key].obj)
    new_value = orig_value + 1e9
    fsi.fs_exp.model_objects[key].obj.value = new_value
    print(f"* orig_value = {orig_value}, new value = {new_value}")

    # try to solve (for real)

    with pytest.raises(RuntimeError) as excinfo:
        fsi.solve()
    print(f"* RuntimeError: {excinfo.value}")


@pytest.mark.unit
def test_has_version():
    fsi = flowsheet_interface()
    d = fsi.dict()
    assert "version" in d
    assert d["version"] > 0


@pytest.mark.unit
def test_to_csv(tmpdir):
    fsi = flowsheet_interface()
    fsi.build()
    outputs = [
        tmpdir / "path.csv",
        str(tmpdir / "filename.csv"),
        open(tmpdir / "fileobj.csv", "w"),
    ]
    for o in outputs:
        num = fsi.fs_exp.to_csv(o)
        assert num > 0
